/* -*- mode: C; c-file-style: "gnu" -*- */
/* Copyright 2004 Red Hat, Inc.
 *
 * This software may be freely redistributed under the terms of the GNU
 * Lesser General Public license.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* This program serves one major purpose:
 *    - Update the fs table in response to HAL events
 *    
 * Additionally, this program offers another option of removing
 * any trace of its previous actions from the fs table.
 *
 * Because it is possible that this program could be invoked multiple
 * times (at bootup or when a user plugs in multiple pieces of
 * hardware), it is very important that all the different instances of
 * this program do not step on each others toes.  Also, it is important
 * that this program does not corrupt the fs table in the event of
 * error conditions like out-of-disk-space and power outages.  
 *
 * For these reasons, all operations are done on temporary copies of
 * /etc/fstab. After any particular instance is done it copies its temporary
 * file over to /etc/fstab.  If another process detects that fstab has
 * changed, then it simply starts over, recopying /etc/fstab and trying again.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#define _GNU_SOURCE
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <fstab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <time.h>
#include <unistd.h>
#include <mntent.h>

#include <popt.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include "libhal/libhal.h"

#define _(a) (a)
#define N_(a) a

#define PROGRAM_NAME "fstab-sync"
#define TEMP_FSTAB_PREFIX ".fstab.hal."
#define TEMP_FSTAB_MAX_LENGTH 64
#define MOUNT_ROOT "/media/"

#define LOCK_TIMEOUT 60          /* seconds */
#define LOCK_TIMEOUT_WAIT 500000 /* microseconds */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE !TRUE
#endif

#define fstab_update_debug(...) if (verbose) fprintf (stderr, __VA_ARGS__)

typedef int boolean;

/** This structure represents either a volume with a mountable filesystem
 *  or a drive that uses media without partition tables.
 *
 *  All fields needs to be non-NULL.
 */
typedef struct
{
  char *udi;             /**< UDI from HAL */
  char *block_device;    /**< Special device file */
  char *type;            /**< Name used in mount point construction, e.g. "cdrw" */
  char *fs_type;         /**< Filesystem type or blank */
  char *mount_point;     /**< Final unique mount point, e.g "/media/cdrw2" */
  char *label;           /**< Label of media or blank */
} Volume;

typedef enum 
{
  FS_TABLE_FIELD_TYPE_BLOCK_DEVICE = 0,
  FS_TABLE_FIELD_TYPE_MOUNT_POINT,
  FS_TABLE_FIELD_TYPE_FILE_SYSTEM_TYPE,
  FS_TABLE_FIELD_TYPE_MOUNT_OPTIONS,
  FS_TABLE_FIELD_TYPE_DUMP_FREQUENCY,
  FS_TABLE_FIELD_TYPE_PASS_NUMBER,
  FS_TABLE_FIELD_TYPE_WHITE_SPACE,
  FS_TABLE_NUM_FIELD_TYPES
} FSTableFieldType;

typedef struct FSTableField
{
  FSTableFieldType type;
  char *value;
  struct FSTableField *next;
} FSTableField;

typedef struct FSTableLine
{
  FSTableField *fields, *tail;

  char *block_device;
  char *device_type;
  char *fs_type;
  char *mount_point;
  char *mount_options;
  char *dump_frequency;
  char *pass_number;

  struct FSTableLine *next;
} FSTableLine;

typedef struct
{
  FSTableLine *lines, *tail;
  char *parse_buffer;
  size_t parse_buffer_length;
  size_t parse_buffer_capacity;
} FSTable;

static LibHalContext *hal_context = NULL;
static boolean verbose = FALSE;
static pid_t pid;

static void fs_table_line_add_field (FSTableLine *line, FSTableField *field);
static boolean fs_table_line_is_generated (FSTableLine *line);
static void fs_table_line_update_pointer (FSTableLine *line, FSTableField *field);

static FSTable *fs_table_new (const char *filename);
static void fs_table_free (FSTable *table);
static boolean fs_table_parse_data (FSTable    *table, 
                                    const char *data, 
                                    size_t      length);


static inline int get_random_int_in_range (int low, int high);
static int open_and_lock_file (const char *filename);
static int open_temp_fstab_file (const char *dir, char **filename);

static char *get_hal_string_property (const char *udi, const char *property);
static boolean mount_device (const char *mount_point);
static boolean udi_is_volume_or_nonpartition_drive (const char *udi);
static Volume *volume_new (const char *udi);
static void volume_determine_device_type (Volume *volume);
static void volume_free (Volume *volume);
static boolean create_mount_point_for_volume (Volume *volume);
static boolean fs_table_add_volume (FSTable *table, Volume *volume);
static FSTableLine *fs_table_remove_volume (FSTable *table, Volume *volume);

static boolean add_udi (const char *udi, boolean should_mount_device);
static boolean remove_udi (const char *udi);
static boolean clean (void);

static FSTableField *
fs_table_field_new (FSTableFieldType type, const char *value)
{
  FSTableField *field;

  field = malloc (sizeof (FSTableLine));

  field->type = type;
  field->value = strdup (value);
  field->next = NULL;

  return field;
}

static void
fs_table_field_free (FSTableField *field)
{

  if (field == NULL)
    return;

  field->type = FS_TABLE_FIELD_TYPE_WHITE_SPACE;

  if (field->value != NULL)
    {
      free (field->value);
      field->value = NULL;
    }
}

static FSTableLine *
fs_table_line_new (void)
{
  FSTableLine *line;

  line = calloc (sizeof (FSTableLine), 1);

  return line;
}

static const char *
get_whitespace (const char *field, int preferred_width)
{
  static const char whitespace[] = "                                           ";
  static const int whitespace_length = (const int) sizeof (whitespace);
  int field_length;

  field_length = strlen (field);

  if (field_length >= preferred_width)
    return " ";

  if (preferred_width > whitespace_length)
    preferred_width = whitespace_length;

  return whitespace + whitespace_length - preferred_width + field_length - 1;
}

static FSTableLine *
fs_table_line_new_from_field_values (const char *block_device,
                                     const char *mount_point,
                                     const char *fs_type,
                                     const char *mount_options,
                                     int dump_frequency,
                                     int pass_number)
{
  FSTableLine *line;
  FSTableField *field;
  char number[32];

  line = fs_table_line_new ();

  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_BLOCK_DEVICE, block_device); 
  fs_table_line_add_field (line, field);

  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_WHITE_SPACE, 
                              get_whitespace (block_device, 24)); 
  fs_table_line_add_field (line, field);

  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_MOUNT_POINT, mount_point); 
  fs_table_line_add_field (line, field);
  
  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_WHITE_SPACE, 
                              get_whitespace (mount_point, 24)); 
  fs_table_line_add_field (line, field);

  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_FILE_SYSTEM_TYPE, fs_type); 
  fs_table_line_add_field (line, field);

  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_WHITE_SPACE, 
                              get_whitespace (fs_type, 8)); 
  fs_table_line_add_field (line, field);

  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_MOUNT_OPTIONS, mount_options);
  fs_table_line_add_field (line, field);

  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_WHITE_SPACE, 
                              get_whitespace (mount_options, 16)); 
  fs_table_line_add_field (line, field);

  snprintf (number, 32, "%d", dump_frequency);
  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_DUMP_FREQUENCY, number); 
  fs_table_line_add_field (line, field);

  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_WHITE_SPACE, 
                              get_whitespace (number, 1)); 
  fs_table_line_add_field (line, field);

  snprintf (number, 32, "%d", pass_number);
  field = fs_table_field_new (FS_TABLE_FIELD_TYPE_PASS_NUMBER, number); 
  fs_table_line_add_field (line, field);

  return line;
}

static void
fs_table_line_add_field (FSTableLine *line, FSTableField *field)
{
  if (line->fields == NULL)
    {
      line->fields = field;
      line->tail = field;
    }
  else
    {
      line->tail->next = field;
      line->tail = field;
    }

  fs_table_line_update_pointer (line, field);
}

static boolean fs_table_line_is_generated (FSTableLine *line)
{

#ifdef USE_NOOP_MOUNT_OPTION
  if (!fs_table_line_has_mount_option (line, "kudzu"))
    return FALSE;
#endif

  if (strncmp (line->mount_point, MOUNT_ROOT, sizeof (MOUNT_ROOT) -1) != 0)
    return FALSE;

  return TRUE;
}

static void
fs_table_line_free (FSTableLine *line)
{
  FSTableField *field;

  if (line == NULL)
    return;

  field = line->fields;
  while (field != NULL)
    {
      FSTableField *this_field;

      this_field = field;
      field = field->next;

      fs_table_field_free (this_field);
    }

  line->fields = NULL;
  line->next = NULL;

  free (line);
}

static void
fs_table_add_line (FSTable *table, FSTableLine *line)
{
  if (table->lines == NULL)
    {
      table->lines = line;
      table->tail = line;
    }
  else
    {
      table->tail->next = line;
      table->tail = line;
    }
}

static char *
fs_table_to_string (FSTable *table, size_t *length)
{
  char *output_string;
  size_t output_string_length, output_string_capacity;
  FSTableLine *line;

  output_string_capacity = 1024;
  output_string = malloc (output_string_capacity);
  output_string[0] = '\0';
  output_string_length = 0;

  line = table->lines;
  while (line != NULL)
    {
      FSTableField *field;

      field = line->fields;
      while (field != NULL)
        {
          size_t field_length;

          field_length = strlen (field->value);

          if (output_string_length + field_length >= output_string_capacity - 1)
            {
              output_string_capacity *= 2;

              output_string = realloc (output_string, output_string_capacity);
            }

          strcpy (output_string + output_string_length, field->value);
          output_string_length += field_length; 

          field = field->next;
        }

      if (output_string_length + 1 >= output_string_capacity - 1)
        {
          output_string_capacity *= 2;

          output_string = realloc (output_string, output_string_capacity);
        }
      output_string[output_string_length++] = '\n';
      output_string[output_string_length] = '\0';

      line = line->next;
    }

  if (length)
    *length = output_string_length;

  return output_string;
}

static FSTable *
fs_table_new (const char *filename)
{
  FSTable *table;
  int input_fd;
  ssize_t bytes_read;
  unsigned char read_buf[1024];

  table = malloc (sizeof (FSTable));

  table->lines = NULL;
  table->parse_buffer_capacity = 128;
  table->parse_buffer = malloc (table->parse_buffer_capacity);
  table->parse_buffer[0] = '\0';
  table->parse_buffer_length = 0;

  input_fd = open_and_lock_file (filename);

  if (input_fd < 0)
    goto error;

  while ((bytes_read = read (input_fd, read_buf, 1024)) != 0)
    {
      if (bytes_read < 0)
        {
          if (errno == EINTR)
            continue;

          fstab_update_debug (_("%d: Could not read from '%s': %s\n"),
                              pid, filename, strerror (errno));
          goto error;
        }

      if (!fs_table_parse_data (table, read_buf, (size_t) bytes_read))
        {
          fstab_update_debug (_("%d: Could not parse data from '%s'\n"), 
			      pid, filename);
          goto error;
        }
    }

  close (input_fd);

  return table;

error:

  fs_table_free (table);

  if (input_fd >= 0)
    close (input_fd);

  return NULL;
}

static boolean
fs_table_write (FSTable *table, int fd)
{
  char *fstab_contents;
  size_t bytes_written, bytes_to_write;

  fstab_contents = fs_table_to_string (table, &bytes_to_write);

  bytes_written = 0;
  while ((bytes_written = write (fd, fstab_contents + bytes_written, 
                                 bytes_to_write)) < bytes_to_write)
    {
      if (bytes_written < 0)
        {
          if (errno == EINTR)
            continue;

          fstab_update_debug (_("%d: Could not write to temporary file: %s\n"),
                              pid, strerror (errno));
          return FALSE;
        }

      bytes_to_write -= bytes_written;
    }

  free (fstab_contents);

  return TRUE;
}

static void
fs_table_free (FSTable *table)
{
  FSTableLine *line;

  line = table->lines;
  while (line != NULL)
    {
      FSTableLine *this_line;

      this_line = line;
      line = line->next;

      fs_table_line_free (this_line);
    }
  table->lines = NULL;

  table->parse_buffer[0] = '\0';
  free (table->parse_buffer);

  table->parse_buffer_length = 0;
  table->parse_buffer_capacity = 0;

  free (table);
}

static void
fs_table_line_update_pointer (FSTableLine *line, FSTableField *field)
{
  switch (field->type)
    {
    case FS_TABLE_FIELD_TYPE_WHITE_SPACE:
      break;
    case FS_TABLE_FIELD_TYPE_BLOCK_DEVICE:
      line->block_device = field->value;
      break;
    case FS_TABLE_FIELD_TYPE_MOUNT_POINT:
      line->mount_point = field->value;
      break;
    case FS_TABLE_FIELD_TYPE_FILE_SYSTEM_TYPE:
      line->fs_type = field->value;
      break;
    case FS_TABLE_FIELD_TYPE_MOUNT_OPTIONS:
      line->mount_options = field->value;
      break;
    case FS_TABLE_FIELD_TYPE_DUMP_FREQUENCY:
      line->dump_frequency = field->value;
      break;
    case FS_TABLE_FIELD_TYPE_PASS_NUMBER:
      line->pass_number = field->value;
      break;
    default:
      assert (FALSE);
      break;
    }
}

static boolean
fs_table_parse_line (FSTable *table, const char *line, size_t length)
{
  FSTableLine *table_line;
  FSTableField *field;
  char *field_value, *p;
  FSTableFieldType current_field;
  size_t i;

  table_line = fs_table_line_new ();

  current_field = FS_TABLE_FIELD_TYPE_BLOCK_DEVICE;
  p = (char *) line;
  i = 0;
  while (*p != '\0' && current_field <= FS_TABLE_FIELD_TYPE_WHITE_SPACE)
    {
      /* First grab the whitespace before the current field 
       */
      while (isspace (line[i]) && i < length)
        i++;

      /* If it's not before the first field or after the last field
       * then we don't want to accept the line ending with whitespace
       */
      if (current_field != FS_TABLE_FIELD_TYPE_WHITE_SPACE
          && (i > length || (line[i] == '\0' && current_field != 0)))
        {
          fstab_update_debug (_("%d: Line ended prematurely\n"), pid);
          return FALSE;
        }

      if (line + i != p)
        {
          field_value = strndup (p, line + i - p);
          field = fs_table_field_new (FS_TABLE_FIELD_TYPE_WHITE_SPACE,
                                      field_value);
          free (field_value);

          fs_table_line_add_field (table_line, field);
        }

      /* Now the actual field ...
       */
      p = (char *) line + i;

      /* If it's a comment, blank line, or end of line, 
       * grab the rest of the line, otherwise just go until whitespace
       */

      if (*p == '#' || 
          current_field == FS_TABLE_FIELD_TYPE_WHITE_SPACE)
        {
          field_value = strndup (p, length - i);
          field = fs_table_field_new (FS_TABLE_FIELD_TYPE_WHITE_SPACE,
                                      field_value);
          fs_table_line_add_field (table_line, field);
          break;
        }

      while (line[i] != '\0' && !isspace (line[i]) && i < length)
        i++;

      if (i > length || (line[i] == '\0' && i < length))
        {

          if (current_field == 0)
            break;
          else
            {
              fstab_update_debug (_("%d: Line ended prematurely\n"), pid);
              return FALSE;
            }
        }

      assert (line + i != p);

      field_value = strndup (p, line + i - p);
      field = fs_table_field_new (current_field, field_value);
      current_field++;
      free (field_value);

      fs_table_line_add_field (table_line, field);

      p = (char *) line + i;
    }

  fs_table_add_line (table, table_line);

  return TRUE;
}

static boolean
fs_table_flush_parse_buffer (FSTable *table)
{
  boolean retval;

  if (table->parse_buffer_length > 0) 
    {
      retval = fs_table_parse_line (table, table->parse_buffer,
                                    table->parse_buffer_length);

      table->parse_buffer[0] = '\0';
      table->parse_buffer_length = 0;

      return retval;
    }

  return TRUE;
}

static boolean 
fs_table_parse_data (FSTable *table, const char *data, size_t length)
{
  boolean retval;
  size_t i;

  for (i = 0; i < length; i++)
    {
      if (data[i] == '\n')
        {
	  /* When a newline is encountered flush the parse buffer so that the
	   * line can be parsed.  Note that completely blank lines won't show
	   * up in the parse buffer, so they get parsed directly.
	   */
	  if (table->parse_buffer_length > 0)
	    retval = fs_table_flush_parse_buffer (table);
	  else
	    retval = fs_table_parse_line (table, "", 1);

          if (retval == FALSE)
            return FALSE;
        }
      else
        {
          if (table->parse_buffer_length + 1 >= 
              table->parse_buffer_capacity - 1)
            {
              table->parse_buffer_capacity *= 2;

              table->parse_buffer = realloc (table->parse_buffer, 
                                             table->parse_buffer_capacity);

            }

          table->parse_buffer[table->parse_buffer_length++] = data[i];
          table->parse_buffer[table->parse_buffer_length] = '\0';
        }
    }

  return TRUE;
}

static inline int
get_random_int_in_range (int low, int high)
{
  return (random () % (high - low)) + low;
}

static int
open_temp_fstab_file (const char *dir, char **filename)
{
  char candidate_filename[TEMP_FSTAB_MAX_LENGTH] = { 0 };
  int fd;

  enum Choice 
    { 
      CHOICE_UPPER_CASE = 0, 
      CHOICE_LOWER_CASE, 
      CHOICE_DIGIT, 
      NUM_CHOICES
    } choice;

  strncpy (candidate_filename, TEMP_FSTAB_PREFIX,
           TEMP_FSTAB_MAX_LENGTH);

  /* Generate a unique candidate filename and open it for writing
   */
  srandom ((unsigned int) time (NULL));
  fd = -1;
  while (fd < 0)
    {
      char *full_path, buf[2] = { 0 };

      if (strlen (candidate_filename) > TEMP_FSTAB_MAX_LENGTH) 
        {
          strncpy (candidate_filename, TEMP_FSTAB_PREFIX,
                   TEMP_FSTAB_MAX_LENGTH);
          candidate_filename[strlen (TEMP_FSTAB_PREFIX)] = '\0';
        }

      choice = (enum Choice) get_random_int_in_range (0, NUM_CHOICES);

      switch (choice)
        {
        case CHOICE_UPPER_CASE:
          buf[0] = (char) get_random_int_in_range ('A' , 'Z' + 1);
          strcat (candidate_filename, buf);
          break;
        case CHOICE_LOWER_CASE:
          buf[0] = (char) get_random_int_in_range ('a' , 'z' + 1);
          strcat (candidate_filename, buf);
          break;
        case CHOICE_DIGIT:
          buf[0] = (char) get_random_int_in_range ('0' , '9' + 1);
          strcat (candidate_filename, buf);
          break;
        default:
          abort ();
          break;
        }

      full_path = (char *) malloc (strlen (dir) + 1 /* / */ + 
                                   strlen (candidate_filename) + 
                                   1 /* \0 */ );
      strcpy (full_path, dir);
      if (full_path[strlen (full_path) - 1] != '/')
        strcat (full_path, "/");
      strcat (full_path, candidate_filename);

      fd = open (full_path, O_CREAT | O_RDWR | O_EXCL, 0644); 

      if (fd < 0) 
        {
          if (errno != EEXIST)
            {
              fstab_update_debug (_("%d: Could not open temporary file for "
                                    "writing in directory '%s': %s\n"),
                                  pid, dir, strerror (errno));
              break;
            }
        }
      else if (fd >= 0 && filename != NULL)
        {
          *filename = strdup (full_path);
          break;
        }

      free (full_path);
    }

  return fd;
}

static boolean
mount_device (const char *mount_point)
{
  char *argv[] = { "/bin/mount", NULL, NULL };
  pid_t mpid;
  int status;

  argv[1] = (char *) mount_point;

  if (!(mpid = fork ()))
    {
      execv (argv[0], argv);
      _exit (1);
    }

  waitpid (mpid, &status, 0);

  if (!WIFEXITED (status) || WEXITSTATUS (status))
    {
      fstab_update_debug (_("%d: /bin/mount failed\n"), pid);
      return FALSE;
    }

  return TRUE;
}

static char *
get_hal_string_property (const char *udi, const char *property)
{
  char *value;

  value = NULL;

  if (hal_device_property_exists (hal_context, udi, property))
    value = hal_device_get_property_string (hal_context, udi, property);

  return value;
}

static boolean
udi_is_volume_or_nonpartition_drive (const char *udi)
{

  if (hal_device_query_capability (hal_context, udi, "volume")) {

    /* we accept volumes with data and known filesystems */

    if (hal_device_property_exists (hal_context, udi, "volume.is_disc.has_data") &&
	!hal_device_get_property_bool (hal_context, udi, "volume.disc.has_data"))
      return FALSE;

    if (hal_device_property_exists (hal_context, udi, "volume.is_filesystem") &&
	!hal_device_get_property_bool (hal_context, udi, "volume.is_filesystem"))
      return FALSE;

    return TRUE;

  } else if (hal_device_query_capability (hal_context, udi, "storage")) {

    /* we accept storage drives with hints that they usually use media
     * without partition tables (such as optical and floppy drives)
     */

    if (hal_device_property_exists (hal_context, udi, "storage.no_partitions_hint") &&
	hal_device_get_property_bool (hal_context, udi, "storage.no_partitions_hint"))
      return TRUE;
  }

  return FALSE;
}

static char *
device_type_normalize (const char *device_type)
{
  char *p, *q, *new_device_type;

  new_device_type = malloc (strlen (device_type) + 1);
  q = new_device_type;
  for (p = (char *) device_type; *p != '\0'; p++)
    {
      if (isspace (*p))
        *q = '_';
      else if (*p == '/')
        *q = '-';
      else
        {
          *q = tolower (*p);

          if (*q < 'a' || *q > 'z')
            *q = '.';
        }
      q++;
    }
  *q = '\0';

  return new_device_type;
}


static char *
compute_cdrom_name (const char *drive_udi)
{
  boolean cdr;
  boolean cdrw;
  boolean dvd;
  boolean dvdplusr;
  boolean dvdplusrw;
  boolean dvdr;
  boolean dvdram;
  char *first;
  char *second;
  char *result;
  
  /* use the capabilities of the optical device */
  
  cdr = hal_device_get_property_bool (hal_context, drive_udi, "storage.cdrom.cdr");
  cdrw = hal_device_get_property_bool (hal_context, drive_udi, "storage.cdrom.cdrw");
  dvd = hal_device_get_property_bool (hal_context, drive_udi, "storage.cdrom.dvd");
  dvdplusr = hal_device_get_property_bool (hal_context, drive_udi, "storage.cdrom.dvdplusr");
  dvdplusrw = hal_device_get_property_bool (hal_context, drive_udi, "storage.cdrom.dvdplusrw");
  dvdr = hal_device_get_property_bool (hal_context, drive_udi, "storage.cdrom.dvdr");
  dvdram = hal_device_get_property_bool (hal_context, drive_udi, "storage.cdrom.dvdram");

  first = NULL;
  if (cdr)
    first = "cdr";
  if (cdrw)
    first = "cdrw";
  
  second = NULL;
  if (dvd)
    second = "dvdrom";
  if (dvdram)
    second = "dvdram";
  if (dvdplusr || dvdr)
    second = "dvdr";
  if (dvdplusrw)
    second = "dvdrw";

  if (first == NULL && second == NULL)
    result = strdup ("cdrom");
  else if (first == NULL)
    result = strdup (second);
  else if (second == NULL)
    result = strdup (first);
  else {
    char buf[64];

    snprintf (buf, 64, "%s_%s", first, second);
    result = strdup (buf);
  }

  return result;
}


static void
volume_determine_device_type (Volume *volume)
{
  char *storage_device_udi;
  char *bus;
  char *drive_type;
  char buf[64];

  volume->type = NULL;

  storage_device_udi = get_hal_string_property (volume->udi,
                                                "block.storage_device");

  if (storage_device_udi == NULL || !hal_device_exists (hal_context, storage_device_udi)) {
    volume->type = strdup ("disk");
    return;
  }

  bus = hal_device_get_property_string (hal_context, storage_device_udi,
					"storage.bus");
  drive_type = hal_device_get_property_string (hal_context, storage_device_udi,
					       "storage.drive_type");

  if (strcmp (drive_type, "cdrom") == 0) {
    
    volume->type = compute_cdrom_name (storage_device_udi);

  } else if (strcmp (drive_type, "disk") == 0) {

    snprintf (buf, 63, "%sdisk", bus);
    volume->type = strdup (buf);

  } else {

    volume->type = strdup (drive_type);

  }
  
  free (drive_type);
  free (bus);
  free (storage_device_udi);
}

static Volume *
volume_new (const char *udi)
{
  Volume *volume;

  if (!udi_is_volume_or_nonpartition_drive (udi))
    return NULL;

  volume = calloc (sizeof (Volume), 1);

  volume->udi = strdup (udi);

  volume->block_device = get_hal_string_property (udi, "block.device");

  if (volume->block_device == NULL)
    {
      volume_free (volume);
      return NULL;
    }

  volume->type = NULL;

  volume_determine_device_type (volume);

  volume->fs_type = get_hal_string_property (udi, "volume.fstype");
  if (volume->fs_type == NULL || strlen (volume->fs_type) == 0)
    {
      volume->fs_type = strdup ("auto");
    }

  volume->label = get_hal_string_property (udi, "volume.label");
  if (volume->label == NULL)
    volume->label = strdup ("");

  return volume;
}

static void
volume_free (Volume *volume)
{

  if (!volume)
    return;

  if (volume->udi != NULL)
    {
      free (volume->udi);
      volume->udi = NULL;
    }

  if (volume->block_device != NULL)
    {
      free (volume->block_device);
      volume->block_device = NULL;
    }

  if (volume->fs_type != NULL)
    {
      free (volume->fs_type);
      volume->fs_type = NULL;
    }

  if (volume->label != NULL)
    {
      free (volume->label);
      volume->fs_type = NULL;
    }

  if (volume->type != NULL)
    {
      free (volume->type);
      volume->type = NULL;
    }
}

/* FIXME: This function should be more fleshed out then it is
 */
static boolean
create_mount_point_for_volume (Volume *volume)
{

  /* FIXME: Should only mkdir if we need to and should do so
   * recursively for each component of the mount root.
   */
  mkdir (MOUNT_ROOT, 0775);

  return (mkdir (volume->mount_point, 0775) != -1) || errno == EEXIST;
}


static boolean
fs_table_has_volume (FSTable *table, Volume *volume)
{
  FSTableLine *line;

  line = table->lines;
  while (line != NULL)
    {
      FSTableField *field;

      field = line->fields;
      while (field != NULL)
        {

          if (field->type == FS_TABLE_FIELD_TYPE_BLOCK_DEVICE)
            {

	      /* Easy, the device file is a match */
              if (strcmp (field->value, volume->block_device) == 0)
                return TRUE;

	      /* Mount by label, more tricky, see below... */
	      if (strncmp (field->value, "LABEL=", 6) == 0 &&
		  strlen (field->value) > 6 &&
		  strcmp (field->value + 6, volume->label) == 0) {
		FSTableField *i;
		char *mount_point;
      
		/* OK, so this new volume has a label that is matched
		 * in the fstab.. Check, via /etc/mtab, whether the
		 * device file for the entry in mtab matches the mount
		 * point 
		 *
		 * (If it's mounted at all, which we assume it
		 * is since no hotpluggable drives should be listed
		 * in /etc/fstab by LABEL. And note that if it's 
		 * not mounted then now we have TWO volumes with
		 * the same label and then everything is FUBAR anyway) 
		 */

		/* first, find the mountpoint from fstab */
		mount_point = NULL;
		for (i = line->fields; i != NULL; i = i->next) {
		  if (i->type == FS_TABLE_FIELD_TYPE_MOUNT_POINT) {
		    mount_point = i->value;
		    break;
		  }
		}

		if (mount_point != NULL) {
		  FILE *f;
		  struct mntent mnt;
		  struct mntent *mnte;
		  char buf[512];
		  char *device_file_from_mount_point;

		  /* good, now lookup in /etc/mtab */
		  device_file_from_mount_point = NULL;
		  if ((f = setmntent ("/etc/mtab", "r")) != NULL) {

		    while ((mnte = getmntent_r (f, &mnt, buf, sizeof(buf))) != NULL) {
		      if (strcmp (mnt.mnt_dir, mount_point) == 0) {
			device_file_from_mount_point = mnt.mnt_fsname;
			break;
		      }
		    }

		    endmntent (f);
		  }

		  /* now see if it's the same device_file as our volume */
		  if (device_file_from_mount_point != NULL &&
		      strcmp (device_file_from_mount_point, 
			      volume->block_device) == 0) {
		    /* Yah it is.. So we're already listed in the fstab */
		    return TRUE;
		  }
		}
	      } /* entry is LABEL= */

            }

          field = field->next;
        }
      line = line->next;
    }

  return FALSE;
}

static boolean
fs_table_add_volume (FSTable *table, Volume *volume)
{
  char *mount_options;
  FSTableLine *line;

  if (fs_table_has_volume (table, volume))
    {
      fstab_update_debug (_("%d: Could not add entry to fstab file: "
                            "block device already listed\n"), pid);
      return FALSE;
    }

    if (hal_device_property_exists (hal_context, volume->udi, "storage.drive_type") &&
	strcmp (hal_device_get_property_string (hal_context, volume->udi, "storage.drive_type"),
		"cdrom") == 0) {
#ifdef USE_NOOP_MOUNT_OPTION
      mount_options = "noauto,user,exec,dev,suid,kudzu,ro";
#else
      mount_options = "noauto,user,exec,dev,suid,ro";
#endif
    } else {
#ifdef USE_NOOP_MOUNT_OPTION
      mount_options = "noauto,user,exec,dev,suid,kudzu";
#else
      mount_options = "noauto,user,exec,dev,suid";
#endif
    }

  line = fs_table_line_new_from_field_values (volume->block_device,
                                              volume->mount_point,
                                              volume->fs_type,
                                              mount_options, 0, 0); 
  fs_table_add_line (table, line);

  return TRUE;
}

static boolean
fs_table_line_has_mount_option (FSTableLine *line, const char *option)
{
  char *p, *q;

  if (line->mount_options == NULL)
    return FALSE;

  p = line->mount_options;
  q = p;
  while (*q != '\0')
    {
      while (*q != ',' && *q != '\0')
        q++;

      if (strncmp (option, p, q - p) == 0)
        return TRUE;

      q++;
      p = q;
    }

  return FALSE;
}

/* Note this function returns the line of the fs table that was
 * removed for convenience (a bit strange, I know)
 */
static FSTableLine *
fs_table_remove_volume (FSTable *table, Volume *volume)
{
  FSTableLine *line, *previous_line; 

  previous_line = NULL;
  line = table->lines;
  while (line != NULL)
    {
      if (line->block_device != NULL
          && strcmp (line->block_device, volume->block_device) == 0
          && fs_table_line_is_generated (line))
        {
          if (previous_line == NULL)
              table->lines = line->next;
          else 
            {
              if (line->next == NULL)
                table->tail = previous_line;
              previous_line->next = line->next;
            }

          return line;
        }

      previous_line = line;
      line = line->next;
    }

  return NULL;
}

static time_t
get_file_modification_time (const char *filename)
{
  struct stat buf = { 0 };

  if (stat (filename, &buf) < 0)
    {
      fstab_update_debug (_("%d: Could not stat '%s': %s\n"),
                          pid, filename, strerror (errno));
    }
  
  return buf.st_mtime;
}

static int
open_and_lock_file (const char *filename)
{
  struct flock lock;
  int fd, saved_errno, lock_status;
  time_t timeout;

  fd = open (filename, O_RDONLY);

  if (fd < 0)
    {
      fstab_update_debug (_("%d: failed to open '%s': %s\n"),
                          pid, filename, strerror (errno));
      return FALSE;
    }

  lock.l_type = F_RDLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start = 0;
  lock.l_len = 0;

  timeout = time (NULL) + LOCK_TIMEOUT;

  while ((lock_status = fcntl (fd, F_SETLK, &lock)) < 0)
    {
      saved_errno = errno;

      if (saved_errno != EACCES 
          && saved_errno != EAGAIN 
          && saved_errno != EINTR)
        break;

      usleep (LOCK_TIMEOUT_WAIT);

      if (time (NULL) > timeout)
        {
          fstab_update_debug (_("%d: timed out waiting for read lock on '%s'\n"),
                              pid, filename);
          close (fd);
          return FALSE;
        }
    }

  if (lock_status < 0)
    {
      fstab_update_debug (_("%d: Could not get read lock on '%s': %s\n"),
                          pid, filename, strerror (saved_errno));
      close (fd);
      return FALSE;
    }

  return fd;
}

static boolean
volume_determine_mount_point (Volume *volume, FSTable *table)
{
  FSTableLine *line;
  char *device_type;
  unsigned long device_number, next_available_device_number = 0;
  long length;
  struct stat buf;

  for (line = table->lines; line != NULL; line = line->next)
    {
      char *p = NULL, *q = NULL, *end = NULL, *mount_root = NULL;

      if (line->mount_point == NULL)
        continue;

      p = strrchr (line->mount_point, '/');

      if (p != NULL)
        {
          mount_root = strndup (line->mount_point, p - line->mount_point + 1);
          p++;
        }
      else
        continue;

      if (strcmp (mount_root, MOUNT_ROOT) != 0)
        {
          free (mount_root);
          continue;
        }
      free (mount_root);

      for (q = p; *q != '\0' && !isspace (*q) && !isdigit (*q); q++);

      device_type = strndup (p, q - p);

      /* If the mount point doesn't end in a number, assume an implied 0 
       */
      if (*q == '\0' || !isdigit (*q))
        device_number = 0;
      else
        device_number = strtoul (q, &end, 10);

      assert (q != end);

      /* Generated mount points don't have numbers in the middle of them--
       * just at the end
       */
      if (end != NULL && *end != '\0')
        {
          free (device_type);
          continue;
        }

      if (strcmp (volume->type, device_type) == 0
          && next_available_device_number == device_number)
        next_available_device_number++;
    }

  length = strlen (volume->type) + sizeof (MOUNT_ROOT) + 6 /* digits */;
  volume->mount_point = malloc (length);

  if (next_available_device_number == 0)
    {
      strcpy (volume->mount_point, MOUNT_ROOT);
      strcat (volume->mount_point, volume->type);
    }
  else if (snprintf (volume->mount_point, length, MOUNT_ROOT"%s%lu",
                     volume->type, next_available_device_number) > length)
    {
      fstab_update_debug (_("%d: Could not use mount point '%s': %s\n"),
                          pid, volume->mount_point, "too long");

      free (volume->mount_point);
      volume->mount_point = NULL;
      return FALSE;
    }

  if (stat (volume->mount_point, &buf) < 0 && errno != ENOENT)
    {
      fstab_update_debug (_("%d: Could not use mount point '%s': %s\n"),
                          pid, volume->mount_point, strerror (errno));

      free (volume->mount_point);
      volume->mount_point = NULL;
      return FALSE;
    }

  return TRUE;
}

static boolean
add_udi (const char *udi, boolean should_mount_device)
{
  Volume *volume;
  FSTable *fs_table;
  char *dir, *last_slash, *temp_filename = NULL;
  time_t fstab_modification_time;
  int fd = -1;

  /* don't add an fstab entry if we were spawned of a device with
   * storage.no_partitions_hint set to TRUE. Per the spec this is
   * exactly when block.no_partitions is TRUE on the volume */
  if (hal_device_query_capability (hal_context, udi, "volume") &&
      hal_device_get_property_bool (hal_context, udi, "block.no_partitions"))
    return FALSE;

  volume = volume_new (udi);

  if (volume == NULL)
    return FALSE;

  dir = strdup (_PATH_FSTAB);
  last_slash = strrchr (dir, '/');
  if (last_slash)
    *last_slash = '\0';

  fs_table = fs_table_new (_PATH_FSTAB);

  if (fs_table == NULL)
    goto error;

  fd = open_temp_fstab_file (dir, &temp_filename);

  if (fd < 0)
    goto error;

  fstab_modification_time = get_file_modification_time (_PATH_FSTAB);

  if (fstab_modification_time == 0)
    goto error;

  if (!volume_determine_mount_point (volume, fs_table))
    goto error;

  if (!fs_table_add_volume (fs_table, volume))
    goto error;

  if (!fs_table_write (fs_table, fd))
    goto error;

  /* Someone changed the fs table under us, better start over.
   */
  if (get_file_modification_time (_PATH_FSTAB) != fstab_modification_time)
    {
      close (fd);
      unlink (temp_filename);
      volume_free (volume);
      return add_udi (udi, should_mount_device);
    }

  if (rename (temp_filename, _PATH_FSTAB) < 0)
    {
      fstab_update_debug (_("%d: Failed to rename '%s' to '%s': %s\n"),
                          pid, temp_filename, _PATH_FSTAB, strerror (errno));
      goto error;
    }

  if (!create_mount_point_for_volume (volume))
    goto error;

  if (should_mount_device && !mount_device (volume->mount_point))
    goto error;

  fstab_update_debug (_("%d: added mount point '%s' for device '%s'\n"),
		      pid, volume->mount_point, volume->block_device);

  close (fd);
  volume_free (volume);

  return TRUE;

error:
  if (fd >= 0)
    close (fd);
  volume_free (volume);
  if (temp_filename != NULL)
    unlink (temp_filename);
  return FALSE;
}

static boolean
remove_udi (const char *udi)
{
  Volume *volume;
  FSTable *fs_table = NULL;
  FSTableLine *line = NULL;
  char *dir, *last_slash, *temp_filename = NULL;
  time_t fstab_modification_time;
  int fd;

  /* don't remove the fstab entry if we were spawned of a device with
   * storage.no_partitions_hint set to TRUE. Per the spec this is
   * exactly when block.no_partitions is TRUE on the volume */
  if (hal_device_query_capability (hal_context, udi, "volume") &&
      hal_device_get_property_bool (hal_context, udi, "block.no_partitions"))
    return FALSE;

  volume = volume_new (udi);

  if (volume == NULL)
    return FALSE;

  dir = strdup (_PATH_FSTAB);
  last_slash = strrchr (dir, '/');
  if (last_slash)
    *last_slash = '\0';

  fs_table = fs_table_new (_PATH_FSTAB);

  fd = open_temp_fstab_file (dir, &temp_filename);

  if (fd < 0)
    goto error;

  fstab_modification_time = get_file_modification_time (_PATH_FSTAB);

  if (fstab_modification_time == 0)
    goto error;

  line = fs_table_remove_volume (fs_table, volume);

  if (line == NULL)
    {
      fstab_update_debug (_("%d: Could not remove device '%s' with UDI '%s' from "
                            "fs table: not found\n"),
                          pid, volume->block_device, udi);
      goto error;
    }

  assert (line->mount_point != NULL);

  if (rmdir (line->mount_point) < 0)
    {
      fstab_update_debug (_("%d: Failed to remove mount point '%s': %s\n"),
                          pid, line->mount_point, strerror (errno));
      goto error;
    }

  fs_table_line_free (line);

  if (!fs_table_write (fs_table, fd))
    goto error;

  /* Someone changed the fs table under us, better start over.
   */
  if (get_file_modification_time (_PATH_FSTAB) != fstab_modification_time)
    {
      close (fd);
      unlink (temp_filename);
      volume_free (volume);
      return remove_udi (udi);
    }

  if (rename (temp_filename, _PATH_FSTAB) < 0)
    {
      fstab_update_debug (_("%d: Failed to rename '%s' to '%s': %s\n"),
                          pid, temp_filename, _PATH_FSTAB, strerror (errno));
      goto error;
    }

  fstab_update_debug (_("%d: removed mount point for device '%s'\n"),
		      pid, volume->block_device);

  close (fd);
  volume_free (volume);

  return TRUE;

error:
  if (fd >= 0)
    close (fd);

  volume_free (volume);

  if (temp_filename != NULL)
    unlink (temp_filename);

  fs_table_line_free (line);

  return FALSE;
}

static void
fs_table_remove_generated_entries (FSTable *table)
{
  FSTableLine *line, *previous_line; 

  previous_line = NULL;
  line = table->lines;
  while (line != NULL)
    {
      if (fs_table_line_is_generated (line))
        {
          if (previous_line == NULL)
              table->lines = line->next;
          else 
            {
              if (line->next == NULL)
                table->tail = previous_line;
              previous_line->next = line->next;
            }
        }

      previous_line = line;
      line = line->next;
    }
}

static boolean
clean (void)
{
  FSTable *fs_table;
  char *dir, *last_slash, *temp_filename = NULL;
  time_t fstab_modification_time;
  int fd;

  dir = strdup (_PATH_FSTAB);
  last_slash = strrchr (dir, '/');
  if (last_slash)
    *last_slash = '\0';

  fs_table = fs_table_new (_PATH_FSTAB);

  fd = open_temp_fstab_file (dir, &temp_filename);

  if (fd < 0)
    goto error;

  fstab_modification_time = get_file_modification_time (_PATH_FSTAB);

  if (fstab_modification_time == 0)
    goto error;

  fs_table_remove_generated_entries (fs_table);

  if (!fs_table_write (fs_table, fd))
    goto error;

  /* Someone changed the fs table under us, better start over.
   */
  if (get_file_modification_time (_PATH_FSTAB) != fstab_modification_time)
    {
      close (fd);
      unlink (temp_filename);
      return clean ();
    }

  if (rename (temp_filename, _PATH_FSTAB) < 0)
    {
      fstab_update_debug (_("%d: Failed to rename '%s' to '%s': %s\n"),
                          pid, temp_filename, _PATH_FSTAB, strerror (errno));
      goto error;
    }

  close (fd);

  return TRUE;

error:
  if (fd >= 0)
    close (fd);
  if (temp_filename != NULL)
    unlink (temp_filename);
  return FALSE;
}

int
main (int argc, const char *argv[])
{
  int i, retval = 0;
  poptContext popt_context;
  boolean should_clean = FALSE, should_mount_device = FALSE;
  char *udi_to_add = NULL, *udi_to_remove = NULL, *hal_device_udi;
  const char **left_over_args = NULL;
  int lockfd = -1;

  pid = getpid ();
  
  struct poptOption options[] = {
      {"add", 'a', POPT_ARG_STRING, &udi_to_add, 0,
        N_("add an entry to fstab"), N_("UDI")},
      {"mount", 'm', POPT_ARG_NONE, &should_mount_device, 0,
        N_("Mount added entry")},
      {"remove", 'r', POPT_ARG_STRING, &udi_to_remove, 0,
        N_("remove an entry from fstab"), N_("UDI")},
      {"clean", 'c', POPT_ARG_NONE, &should_clean, 0,
        N_("Remove all generated entries from fstab")},
      {"verbose", 'v', POPT_ARG_NONE, &verbose, 0,
        N_("Report detailed information about operation progress")},

      POPT_AUTOHELP

      {NULL, '\0', 0, NULL},
  };

  popt_context = poptGetContext (PROGRAM_NAME, argc, argv, options, 0);

  while ((i = poptGetNextOpt (popt_context)) != -1)
    {
      if (i < -1)
        {
          poptPrintHelp (popt_context, stderr, 0);
          return 1;
        }
    }

  /* accept "add" and "remove" / HAL environment variables 
   * in addition to "--add" and "--remove" above
   */
  left_over_args = poptGetArgs (popt_context);

  hal_device_udi = getenv ("UDI");
  if (hal_device_udi != NULL) {
    char *caps;

    /* when we are invoked by hald, make some early tests using the
     * exported environment so we don't have to connect to hald */

    caps = getenv ("HAL_PROP_INFO_CAPABILITIES");

    /* if there are no info.capabilities just bail out */
    if (caps == NULL) {
      retval = 0;
      goto out;
    }

    /* we only handle hal device objects of capability 'volume' or 'storage' */
    if (caps != NULL) {

      if (strstr (caps, "volume") == NULL &&
	  strstr (caps, "storage") == NULL) {
	retval = 0;
	goto out;
      }
    }


    fstab_update_debug (_("%d: ###################################\n"), pid);
    fstab_update_debug (_("%d: %s entering; %s udi=%s\n"), 
			pid, PROGRAM_NAME, argv[1], hal_device_udi);
    
    lockfd = open (_PATH_FSTAB, O_RDONLY);
    if (lockfd < 0) {
      fstab_update_debug (_("%d: couldn't open %s O_RDONLY; bailing out\n"), 
			  pid, _PATH_FSTAB);
      retval = 1;
      goto out;
    }
    fstab_update_debug (_("%d: Acquiring advisory lock on " 
			  _PATH_FSTAB "\n"), pid);
    if (flock (lockfd, LOCK_EX) != 0) {
      fstab_update_debug (_("%d: Error acquiring lock '%s'; bailing out\n"),
			    pid, strerror(errno));
      retval = 1;
      goto out;
    }
    fstab_update_debug (_("%d: Lock acquired\n"), pid);
  }
  

  if (left_over_args)
  for (i = 0; left_over_args[i] != NULL; i++)
    {
      if (strcmp (left_over_args[i], "add") == 0 && udi_to_add == NULL)
        {
          udi_to_add = strdup (hal_device_udi);
          break;
        }
      
      if (strcmp (left_over_args[i], "remove") == 0)
        {
          udi_to_remove = strdup (hal_device_udi);
          break;
        }
    }
  poptFreeContext (popt_context);

  if (udi_to_add || udi_to_remove)
    {
      LibHalFunctions halFunctions = { NULL };

      hal_context = hal_initialize (&halFunctions, 0);

      if (udi_to_add)
        retval = !add_udi (udi_to_add, should_mount_device);

      if (udi_to_remove)
        retval |= !remove_udi (udi_to_remove);

      hal_shutdown (hal_context);
    }
  else if (should_clean)
    retval = clean ();
  else
    {
      poptPrintHelp (popt_context, stderr, 0);
      return 1;
    }

  if (hal_device_udi != NULL) {

    fstab_update_debug (_("%d: Releasing advisory lock on %s\n"), 
			pid, _PATH_FSTAB);
    if (flock (lockfd, LOCK_EX) != 0) {
      fstab_update_debug (_("%d: Error releasing lock '%s'\n"), pid, 
	      strerror(errno));
      retval = 1;
    } else {
      fstab_update_debug (_("%d: Lock released\n"), pid);
    }
    close (lockfd);

    fstab_update_debug (_("%d: %s exiting; %s udi=%s\n"), 
			pid, PROGRAM_NAME, argv[1], hal_device_udi);

    fstab_update_debug (_("%d: ###################################\n"), pid);
    fstab_update_debug (_("\n"));
  }

out:
  return retval;
}
